#include "config.h"

#if HAVE_AFINET6
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <linux/route.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <resolv.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "version.h"
#include "net-support.h"
#include "pathnames.h"
#define  EXTERN
#include "net-locale.h"

#include "net-features.h"

extern struct aftype inet6_aftype;

/* print info from /proc */
int rprint_fib6(int ext, int numeric)
{
	char buff[4096], iface[16], flags[16];
	char gate_addr[128], net_addr[128];
	unsigned int refcnt, use, metric, preflen, mtu, window, irtt, iflags;
	int num;
	struct sockaddr_in6 snet, sgate; 

	FILE *fp = fopen(_PATH_PROCNET_ROUTE6, "r");

	if (!fp) {
		ESYSNOT("getroute", "INET6 FIB");
		return 1;
	}
	printf(NLS_CATGETS(catfd, inetSet, inet_table, "Kernel IPv6 routing table\n"));

	/* ext represents the -e options (none=1, -e=2, -ee=3) */
	switch (ext) 
	{
	case 1: 
		printf(NLS_CATGETS(catfd, inetSet, inet_header1,
				   "Destination/Prefix\t\t\t\t   Gateway\t\t\t\t\t"
				   "  Flags Metric RefCnt Use   Iface\n"));
		break;

	case 2:
		printf(NLS_CATGETS(catfd, inetSet, inet_header2,
				   "Destination/Prefix\t\t\t\t   Gateway\t\t\t\t\t"
				   "  Flags MTU   Window IRTT  Iface\n"));
		break;

	case 3: 
		printf(NLS_CATGETS(catfd, inetSet, inet_header3,
				   "Destination/Prefix\t\t\t\t   Gateway\t\t\t\t\t"
				   "  Flags Metric RefCnt Use   Iface "
				   "MTU   Window IRTT\n"));
		break;
	}

	fgets(buff, 1023, fp);	/* Skip first line */

	while (fgets(buff, 1023, fp)) {
		num = sscanf(buff, "%s\t%s\t%s\t%04x\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
			     iface,
			     net_addr,
			     gate_addr,
			     &iflags,
			     &refcnt,
			     &use,
			     &metric,
			     &preflen,
			     &mtu,
			     &window,
			     &irtt);

		if (num < 11)
			continue;

		/* Fetch and resolve the target address. */
		/* use field sin6_port to store prefix length (as argument for inet6_aftype.sprint) */
		snet.sin6_port = preflen;
		(void) inet6_aftype.input(FROM_FIB, net_addr, (struct sockaddr *)&snet);
		sprintf(net_addr,  "%s/%d", inet6_aftype.sprint((struct sockaddr *)&snet, (numeric | FLAG_TARGET)), preflen);

		/* Fetch and resolve the gateway address. */
		(void) inet6_aftype.input(FROM_FIB, gate_addr, (struct sockaddr *)&sgate);
		sprintf(gate_addr, "%s", inet6_aftype.sprint((struct sockaddr *)&sgate, numeric));

		/* Decode the flags. */
		flags[0] = '\0';
		if (iflags & RTF_UP)
			strcat(flags, "U");
		if (iflags & RTF_GATEWAY)
			strcat(flags, "G");
		if (iflags & RTF_HOST)
			strcat(flags, "H");
		if (iflags & RTF_REINSTATE)
			strcat(flags, "R");
		if (iflags & RTF_DYNAMIC)
			strcat(flags, "D");
		if (iflags & RTF_MODIFIED)
			strcat(flags, "M");
		if (iflags & RTF_DCACHE)
			strcat(flags, "C");
			
#if HAVE_RTF_REJECT
		if (iflags & RTF_REJECT)
			strcpy(flags, "!");
#endif

		/* Print the info. */
		if (ext == 1) {
#if HAVE_RTF_REJECT
			if (iflags & RTF_REJECT)
			  printf("%-50s %-46s %-5s %-6u %-6s %-5u %-4s\n", net_addr, "-", flags, metric, "-", use, "-");
			else
#endif
			  printf("%-50s %-46s %-5s %-6u %-6u %-5u %-4s\n", net_addr, gate_addr, flags, metric, refcnt, use, iface);
		}
		if (ext == 2) {
#if HAVE_RTF_REJECT
			if (iflags & RTF_REJECT)
			  printf("%-50s %-46s %-5s %-5s %-6s %-5s %-4s\n", net_addr, "-", flags, "-", "-", "-", "-");
			else
#endif
			  printf("%-50s %-46s %-5s %-5u %-6u %-5u %-4s\n", net_addr, gate_addr, flags, mtu, window, irtt, iface);
		}
		if (ext >= 3) {
#if HAVE_RTF_REJECT
			if (iflags & RTF_REJECT)
			  printf("%-50s %-46s %-5s %-6u %-6s %-5u %-5s %-5s %-6s %-5s\n", net_addr, gate_addr, flags, metric, "-", use, "-", "-", "-", "-");
			else
#endif
			  printf("%-50s %-46s %-5s %-6u %-6u %-5u %-5s %-5u %-6u %-5u\n", net_addr, gate_addr, flags, metric, refcnt, use, iface, mtu, window, irtt);
		}
	}

	(void) fclose(fp);
	return (0);
}

int rprint_cache6(int ext, int numeric)
{
	FILE *fp = fopen(_PATH_PROCNET_RTCACHE, "r");

	if (!fp) {
		ESYSNOT("getroute", "INET CACHE");
		return 1;
	}
	printf(NLS_CATGETS(catfd, inetSet, inet_cache, "Kernel IP routing cache\n"));
	fprintf(stderr, "INET CACHE: This needs to be written!\n");

	(void) fclose(fp);
	return (0);
}

int INET6_rprint(int options)
{
	int ext = options & FLAG_EXT;
	int numeric = options & (FLAG_NUM | FLAG_SYM);
	int rc = E_INTERN;

	if (options & FLAG_FIB)
		if ((rc = rprint_fib6(ext, numeric)))
			return (rc);
	if (options & FLAG_CACHE)
		rc = rprint_cache6(ext, numeric);

	return (rc);
}

#endif				/* HAVE_AFINET6 */


