#include "config.h"

#if HAVE_AFINET6
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <linux/route.h>
#include <linux/route6.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <resolv.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "version.h"
#include "net-support.h"
#include "pathnames.h"
#define  EXTERN
#include "net-locale.h"

#include "net-features.h"

extern struct aftype inet6_aftype;

static int skfd = -1;

static int usage(void)
{
	fprintf(stderr, "Usage: inet6_route [-vF] del [::]Target/PrefixLength [gw Gw] [metric M] [[dev] If]\n");
	fprintf(stderr, "       inet6_route [-vF] add [::]Target/PrefixLength [gw Gw] [metric M]\n");
	fprintf(stderr, "                              [mss Mss] [window W] [irtt I]\n");
	fprintf(stderr, "                              [mod] [dyn] [reinstate] [[dev] If]\n");
	fprintf(stderr, "       inet6_route [-vF4] add [::]Target/PrefixLength [metric M] reject\n");
	fprintf(stderr, "       inet6_route [-FC] flush      NOT supported\n");
	return (E_USAGE);
}


static void KeepAddressPrefix (struct in6_addr * sin6, unsigned int preflen)
{
unsigned int NbWordErased, i;
__u32 shift;

	NbWordErased = (IN6ADDRSZ*8 - preflen) / 32;
	for (i= NbWordErased ; i; i--) {
		sin6->s6_addr32[4-i] = 0;
	}
	if (preflen/32 + NbWordErased < 4) {
		i = 32 - (preflen % 32);
		shift = ntohl( sin6->s6_addr32[4 - NbWordErased -1]);
		shift >>= i;
		shift <<= i;
		sin6->s6_addr32[4 - NbWordErased -1] = htonl(shift);
	}
		
}

static int INET6_setroute(int action, int options, char **args)
{
        struct in6_rtmsg rt;
	char target[128], gateway[128] = "NONE", *pref;
	struct sockaddr_in6 sin6;
	int isnet, preflen, inputmode;

	/* Clean out the RTREQ structure. */
	memset((char *) &rt, 0, sizeof(struct in6_rtmsg));

	if (*args == NULL)
		return (usage());

	strcpy(target, *args++);

	if (strcmp(target, "default")) {
		/* Separate adress from prefix */
		if ((pref = memchr(target, '/', strlen(target)))) {
			*pref = 0;
			pref++;		/* begin of prefix number */
			preflen = strtol(pref, &pref, 10);
			if ((pref) && (*pref != 0))
				return (usage());
			if ((preflen<0) || (preflen>128))
				return (usage());
		} else
			preflen = 128;		
	} else {
		preflen = 0;	/* default route */
	}
	rt.rtmsg_prefixlen = preflen;

	if (options & FLAG_IP4COMP) 
		inputmode = FROM_DNS_4;
	else
		inputmode = FROM_DNS;

	if ((isnet = inet6_aftype.input(inputmode, target, (struct sockaddr *) &sin6)) < 0) {
	  	inet6_aftype.herror(target);
		return (1);
	}
	
	if ((preflen < 128) && (preflen >0)) {
		isnet = NET_ROUTE;
		KeepAddressPrefix (&sin6.sin6_addr, preflen);
	}
		
	COPY_ADDR6(rt.rtmsg_dst, sin6.sin6_addr);

	/* Fill in the other fields. */
	rt.rtmsg_flags = (RTF_UP | RTF_HOST);
	if (isnet)
		rt.rtmsg_flags &= ~RTF_HOST;

	while (*args) {
		if (!strcmp(*args, "metric")) {
			int metric;

			args++;
			if (!*args || !isdigit(**args))
				return (usage());
			metric = atoi(*args);
#if HAVE_NEW_ADDRT
			rt.rtmsg_metric = metric;
#else
			ENOSUPP("inet_setroute", "NEW_ADDRT (metric)");
#endif
			args++;
			continue;
		}
		if (!strcmp(*args, "gw") || !strcmp(*args, "gateway")) {
			args++;
			if (!*args)
				return (usage());
			if (rt.rtmsg_flags & RTF_GATEWAY)
				return (usage());
			strcpy(gateway, *args);
			if ((isnet = inet6_aftype.input(inputmode, gateway, (struct sockaddr *)&sin6)) < 0) {
				inet6_aftype.herror(gateway);
				return (E_LOOKUP);
			}
			if (isnet) {
				fprintf(stderr, NLS_CATGETS(catfd, routeSet, route_cant_use,
							    "route: %s: cannot use a NETWORK as gateway!\n"),
					gateway);
				return (E_OPTERR);
			}
			COPY_ADDR6(rt.rtmsg_gateway, sin6.sin6_addr);
			rt.rtmsg_flags |= RTF_GATEWAY;
			args++;
			continue;
		}
		if (!strcmp(*args, "mss")) {
			args++;
			rt.rtmsg_flags |= RTF_MSS;
			if (!*args)
				return (usage());
			rt.rtmsg_mss = atoi(*args);
			args++;
			if (rt.rtmsg_mss < 64 || rt.rtmsg_mss > 32768) {
				fprintf(stderr, NLS_CATGETS(catfd, routeSet, route_MSS, "route: Invalid MSS.\n"));
				return (E_OPTERR);
			}
			continue;
		}
		if (!strcmp(*args, "window")) {
			args++;
			if (!*args)
				return (usage());
			rt.rtmsg_flags |= RTF_WINDOW;
			rt.rtmsg_window = atoi(*args);
			args++;
			if (rt.rtmsg_window < 128 || rt.rtmsg_window > 32768) {
				fprintf(stderr, NLS_CATGETS(catfd, routeSet, route_window, "route: Invalid window.\n"));
				return (E_OPTERR);
			}
			continue;
		}
/* FIXME ??? */
#if 0
		if (!strcmp(*args, "irtt")) {
			args++;
			if (!*args)
				return (usage());
			args++;
#if HAVE_RTF_IRTT
			rt.rtmsg_flags |= RTF_IRTT;
			rt.rtmsg_irtt = atoi(*(args - 1));
			rt.rtmsg_irtt *= (HZ / 100);	/* FIXME */
#if 0				/* FIXME: do we need to check anything of this? */
			if (rt.rtmsg_irtt < 1 || rt.rtmsg_irtt > (120 * HZ)) {
				fprintf(stderr, NLS_CATGETS(catfd, routeSet, route_irtt, "route: Invalid initial rtt.\n"));
				return (E_OPTERR);
			}
#endif
#else
			ENOSUPP("inet_setroute", "RTF_IRTT");
#endif
			continue;
		}
#endif /* 0 */
		if (!strcmp(*args, "reject")) {
			args++;
#if HAVE_RTF_REJECT
			rt.rtmsg_flags |= RTF_REJECT;
#else
			ENOSUPP("inet_setroute", "RTF_REJECT");
#endif
			continue;
		}
		if (!strcmp(*args, "mod")) {
			args++;
			rt.rtmsg_flags |= RTF_MODIFIED;
			continue;
		}
		if (!strcmp(*args, "dyn")) {
			args++;
			rt.rtmsg_flags |= RTF_DYNAMIC;
			continue;
		}
		if (!strcmp(*args, "reinstate")) {
			args++;
			rt.rtmsg_flags |= RTF_REINSTATE;
			continue;
		}
		if (!strcmp(*args, "device") || !strcmp(*args, "dev")) {
			args++;
			if (!*args)
				return (usage());
		} else if (args[1])
			return (usage());

		strcpy(rt.rtmsg_device, *args);
		args++;
	}

#if HAVE_RTF_REJECT
	if ((rt.rtmsg_flags & RTF_REJECT) && !rt.rtmsg_device)
		strcpy(rt.rtmsg_device, "lo");
#endif

/* FIXME  If prefix length is not fixed, then
   it is a host, and prefix is then fixed to 128 */
#if 0
	/* Fill out netmask if still unset */
	if ((action == RTACTION_ADD) && rt.rtmsg_flags & RTF_HOST)
		mask_in_addr(rt) = 0xffffffff;
#endif
	/* Create a socket to the INET kernel. */
	if ((skfd = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		return (E_SOCK);
	}
	/* Tell the kernel to accept this route. */
	if (action == RTACTION_DEL) {
		rt.rtmsg_type = SIOCDELRT;
		if (ioctl(skfd, SIOCDELRT, &rt) < 0) {
			perror("SIOCDELRT");
			close(skfd);
			return (E_SOCK);
		}
	} else {
		rt.rtmsg_type = SIOCADDRT;
		if (ioctl(skfd, SIOCADDRT, &rt) < 0) {
			perror("SIOCADDRT");
			close(skfd);
			return (E_SOCK);
		}
	}

	/* Close the socket. */
	(void) close(skfd);
	return (0);
}

int INET6_rinput(int action, int options, char **args)
{
	if (action == RTACTION_FLUSH) {
		fprintf(stderr, "Flushing `inet6' routing table not supported\n");
		return (usage());
	}
	if (options & FLAG_CACHE) {
		fprintf(stderr, "Modifying `inet6' routing cache not supported\n");
		return (usage());
	}
	if ((*args == NULL) || (action == RTACTION_HELP))
		return (usage());

	return (INET6_setroute(action, options, args));
}
#endif				/* HAVE_AFINET6 */
